from libpalette.command import CommandId

class PaletteError(Exception):
    pass

class CommandNotFoundError(PaletteError):
    def __init__(self, identifier: CommandId) -> None:
        super().__init__(f'Command with identifier {identifier} does not exist')

class QuitException(Exception):
    def __init__(self) -> None:
        super().__init__('Quitting...')



import os
from string import whitespace as WHITESPACE

def try_remove_file(path: str) -> None:
    try:
        os.unlink(path)
    except OSError:
        if os.path.exists(path):
            print(f'Error: could not unlink file at path {path}')

def remove_whitespace(string: str) -> str:
    return string.translate(str.maketrans('', '', WHITESPACE))



import datetime
import subprocess
from typing import Callable, List

from libpalette.rofi import Rofi
from libpalette.command import CommandId
from libpalette.connection import Connection
from libpalette.configuration import Configuration
from libpalette.commands_manager import CommandsManager
from libpalette.keybindings_manager import KeybindingsManager
from libpalette.exceptions import PaletteError, QuitException, CommandNotFoundError

class ConnectionHandler:
    def handle_connection(self, connection: Connection) -> None:
        raise NotImplementedError

HELP_MESSAGE = """\
usage: palette COMMAND

Controls the Palette daemon by communication with it through a socket.

commands:
    help   - Show this help message and exit.
    show   - Show the command palette.
    quit   - Ask paletted to suicide.
    reload - Reload the commands json file. 
             If a keybindings backend is running, 
             the keybindings will also be updated accordingly.
"""

class PaletteConnectionHandler(ConnectionHandler):
    _commands_manager: CommandsManager
    _keybindings_manager: KeybindingsManager
    _rofi: Rofi
    _configuration: Configuration

    def __init__(self, commands_manager: CommandsManager, keybindings_manager: KeybindingsManager, configuration: Configuration) -> None:
        self._commands_manager = commands_manager
        self._keybindings_manager = keybindings_manager
        self._rofi = Rofi(configuration)
        self._configuration = configuration
    
    def _handle_no_arguments(self, args: List[str], connection: Connection) -> None:
        connection.send_error('No arguments given')
    
    def _handle_unknown_command(self, args: List[str], connection: Connection) -> None:
        connection.send_error(f'Unknown command {repr(args[0])}')
    
    def _execute_command(self, identifier: CommandId) -> None:
        command = self._commands_manager.get_command(identifier)
        if command is None:
            raise CommandNotFoundError(identifier)
        command.last_used = datetime.datetime.now()
        subprocess.Popen(['sh', '-c', command.exec_string])
    
    def _handle_show(self, args: List[str], connection: Connection) -> None:
        self._rofi.show_palette(self._commands_manager.get_commands())
        
    def _handle_execute(self, args: List[str], connection: Connection) -> None:
        if len(args) == 1:
            connection.send_error('Invalid number of (expected 1)')
            return
        try:
            command_id = CommandId(args[1])
        except ValueError:
            connection.send_error(f'{repr(args[0])} is not a valid command ID')
            return
        self._execute_command(command_id)
    
    def _handle_help(self, args: List[str], connection: Connection) -> None:
        if len(args) == 1:
            connection.send(HELP_MESSAGE)
        else:
            connection.send_error('Invalid arguments (expected 0)')

    def _handle_reload(self, args: List[str], connection: Connection) -> None:
        self._commands_manager.reload()
        self._keybindings_manager.load(self._commands_manager.get_commands().values())
    
    def _handle_quit(self, args: List[str], connection: Connection) -> None:
        raise QuitException

    def _get_handler(self, args: List[str]) -> Callable[[List[str], Connection], None]:
        return self._handle_no_arguments if len(args) == 0 else {
            'show': self._handle_show,
            'help': self._handle_help,
            'quit': self._handle_quit,
            'reload': self._handle_reload,
            '__execute': self._handle_execute, # needed for internal use
        }.get(args[0]) or self._handle_unknown_command
    
    def handle_connection(self, connection: Connection) -> None:
        args = connection.recv_args()
        if self._configuration.verbose:
            print('Args:', args)
        try:
            self._get_handler(args)(args, connection)
        except PaletteError as e:
            connection.send_error(str(e))



import random
import datetime
from dataclasses import dataclass
from typing import Optional, Union, Set

COMMAND_ID_SIZE = 4
COMMAND_ID_BASE = 10
MAX_COMMAND_ID = 10 ** COMMAND_ID_SIZE

class CommandId:
    value: int

    def __init__(self, value: Union[int, str]) -> None:
        self.value = value if isinstance(value, int) else int(value, COMMAND_ID_BASE)
    
    def __str__(self) -> str:
        return str(self.value).zfill(COMMAND_ID_SIZE)
    
    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.value == other.value
    
class CommandIdFactory:
    _existing_ids: Set[int]

    def __init__(self) -> None:
        self._existing_ids = set()
    
    def _get_unused_id(self) -> int:
        while True:
            r = random.randint(0, MAX_COMMAND_ID)
            if r not in self._existing_ids:
                return r
    
    def create(self, value: Optional[Union[int, str]] = None) -> CommandId:
        if value is None:
            value = self._get_unused_id()
        assert value not in self._existing_ids
        new_id = CommandId(value)
        self._existing_ids.add(new_id.value)
        return new_id

@dataclass
class Command:
    description: str
    exec_string: str
    identifier: CommandId
    keybinding: Optional[str] = None
    metadata: Optional[str] = None # invisible search terms
    last_used: datetime.datetime = datetime.datetime.min

class CommandFactory:
    _id_factory: CommandIdFactory

    def __init__(self, id_factory: CommandIdFactory) -> None:
        self._id_factory = id_factory
    
    def create(self, description: str, keybinding: Optional[str], exec_string: str, metadata: Optional[str], identifier: Optional[Union[int, str]] = None) -> Command:
        return Command(
            description,
            exec_string,
            self._id_factory.create(identifier),
            keybinding,
            metadata
        )



from typing import Iterable

from libpalette.command import Command
from libpalette.keybindings_manager import KeybindingsManager

class KeybindingsManagerMock(KeybindingsManager):
    def load(self, keybindings: Iterable[Command]) -> None:
        pass

    def __enter__(self) -> 'KeybindingsManagerMock':
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        pass



import socket
from typing import Union, List

class Connection:
    _socket: socket.socket
    _closed: bool

    def __init__(self, connection_socket: socket.socket) -> None:
        self._socket = connection_socket
        self._closed = False

    def __enter__(self) -> 'Connection':
        return self
    
    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        if not self._closed:
            self.close()

    def close(self) -> None:
        self._socket.close()
        self._closed = True

    def send(self, message: Union[str, bytes]) -> None:
        self._socket.sendall(f'{message}\n'.encode() if isinstance(message, str) else message)
    
    def send_error(self, message: Union[str, bytes]) -> None:
        self.send(b'\x07' + (f'{message}\n'.encode() if isinstance(message, str) else message))
    
    def recv(self, buffsize: int = 4096) -> bytes:
        return self._socket.recv(buffsize)
    
    def recv_args(self) -> List[str]:
        return [arg.decode() for arg in self.recv().split(b'\0') if arg]



from typing import Iterable

from libpalette.command import Command

class KeybindingsManager:
    def load(self, keybindings: Iterable[Command]) -> None:
        raise NotImplementedError

    def __enter__(self) -> 'KeybindingsManager':
        raise NotImplementedError

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        raise NotImplementedError



from signal import SIGUSR1
from typing import Optional
from subprocess import Popen

class SxhkdProcess:
    _process: Popen

    """
    """
    def __init__(self, *, count:        Optional[int] = None,
                          timeout:      Optional[int] = None,
                          config_file:  Optional[str] = None,
                          redir_file:   Optional[str] = None,
                          status_fifo:  Optional[str] = None,
                          abort_keysym: Optional[str] = None) -> None:
        args = ['sxhkd']
        if count        is not None: args += ['-m', str(count)  ]
        if timeout      is not None: args += ['-t', str(timeout)]
        if config_file  is not None: args += ['-c', config_file ]
        if redir_file   is not None: args += ['-r', redir_file  ]
        if status_fifo  is not None: args += ['-s', status_fifo ]
        if abort_keysym is not None: args += ['-a', abort_keysym]
        self._process = Popen(args)
    
    def reload_config(self) -> None:
        self._process.send_signal(SIGUSR1)
    
    def kill(self) -> None:
        self._process.kill()


from socket import socket
from typing import NamedTuple
from argparse import ArgumentParser

from libpalette.sxhkd import Sxhkd
from libpalette.version import VERSION
from libpalette.server import UnixSocketServer
from libpalette.exceptions import QuitException
from libpalette.configuration import Configuration
from libpalette.commands_manager import CommandsManager
from libpalette.connection_handler import ConnectionHandler
from libpalette.keybindings_manager import KeybindingsManager
from libpalette.connection_handler import PaletteConnectionHandler
from libpalette.keybindings_manager_mock import KeybindingsManagerMock

def parse_arguments() -> Configuration:
    DEFAULT_CONFIGURATION = Configuration('')

    parser = ArgumentParser(description='Daemon process for palette')
    parser.add_argument(
        '-c', '--commands',
        type = str,
        required = True,
        help = 'Commands json file path.'
    )
    parser.add_argument(
        '-b', '--backend',
        type = str,
        required = True,
        default = DEFAULT_CONFIGURATION.keybindings_backend,
        choices = ['none', 'sxhkd'],
        help = f'Backend to use for keybindings management (default is {DEFAULT_CONFIGURATION.keybindings_backend})'
    )
    parser.add_argument(
        '-v', '--version',
        action = 'version',
        version = VERSION,
        help = 'Show version number and exit'
    )
    parser.add_argument(
        '-V', '--verbose',
        action = 'store_true',
        help = 'Enable verbose output'
    )
    parser.add_argument(
        '-s', '--socket',
        type = str,
        default = DEFAULT_CONFIGURATION.socket_path,
        help = f'Socket to listen for incoming connections (default is {DEFAULT_CONFIGURATION.socket_path})'
    )
    args = parser.parse_args()
    return Configuration(
        commands_path = args.commands,
        keybindings_backend = args.backend,
        verbose = args.verbose,
        socket_path = args.socket
    )

def get_keybindings_manager(keybindings_backend: str) -> KeybindingsManager:
    if keybindings_backend == 'none':
        return KeybindingsManagerMock()
    if keybindings_backend == 'sxhkd':
        return Sxhkd('/tmp/palette_sxhkdrc')
    raise Exception('Unreachable')

def run():
    configuration = parse_arguments()

    commands_manager = CommandsManager(configuration.commands_path)
    keybindings_manager = get_keybindings_manager(configuration.keybindings_backend)
    keybindings_manager.load(commands_manager.get_commands().values())
    with keybindings_manager:
        connection_handler = PaletteConnectionHandler(commands_manager, keybindings_manager, configuration)
        with UnixSocketServer(configuration.socket_path, connection_handler) as server:
            server.run()

def main() -> int:
    try:
        run()
    except QuitException:
        return 0
    return 1

if __name__ == '__main__':
    main()



from typing import Iterable, Optional, overload

from libpalette.command import Command
from libpalette.utils import try_remove_file
from libpalette.sxhkd_process import SxhkdProcess
from libpalette.configuration import Configuration
from libpalette.keybindings_manager import KeybindingsManager

class Sxhkd(KeybindingsManager):
    _commands: Iterable[Command]
    _sxhkdrc_path: str
    _process: Optional[SxhkdProcess]

    def __init__(self, sxhkdrc_path: str) -> None:
        self._sxhkdrc_path = sxhkdrc_path
        self._commands = []
        self._process = None
    
    def _generate_sxhkdrc(self) -> str:
        return ''.join(
            f'# {c.description}\n{c.keybinding}\n    {c.exec_string}\n\n'
            for c in self._commands if c.keybinding is not None
        )

    def _dump_sxhkdrc(self) -> None:
        with open(self._sxhkdrc_path, 'w') as sxhkdrc_file:
            sxhkdrc_file.write(self._generate_sxhkdrc())
    
    def load(self, commands: Iterable[Command]) -> None:
        self._commands = commands
        if self._process is not None:
            self._dump_sxhkdrc()
            self._process.reload_config()
    
    def __enter__(self) -> KeybindingsManager:
        self._dump_sxhkdrc()
        self._process = SxhkdProcess(config_file = self._sxhkdrc_path)
        return self
        
    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        assert self._process is not None
        self._process.kill()
        self._process = None
        try_remove_file(self._sxhkdrc_path)



import json
from typing import Dict, Optional, Iterable

from libpalette.command import Command, CommandId, CommandFactory, CommandIdFactory

class CommandsManager:
    _commands_json_path: str
    _command_factory: CommandFactory
    _commands: Dict[CommandId, Command]

    def __init__(self, commands_json_path: str) -> None:
        self._commands_json_path = commands_json_path
        self.reload()
    
    def add_command(self, command: Command) -> None:
        self._commands[command.identifier] = command
    
    def get_command(self, identifier: CommandId) -> Optional[Command]:
        return self._commands.get(identifier)
    
    def get_commands(self) -> Dict[CommandId, Command]:
        return self._commands
    
    @staticmethod
    def _command_to_dict(command: Command) -> Dict[str, str]:
        result = {
            "description": command.description,
            "exec": command.exec_string,
        }
        if command.keybinding:
            result["keybinding"] = command.keybinding
        return result
    
    def store_commands(self) -> None:
        commands_list = list(map(CommandsManager._command_to_dict, self._commands.values()))
        with open(self._commands_json_path, 'w') as file:
            json.dump(commands_list, file, indent=4)
    
    def reload(self) -> None:
        self._command_factory = CommandFactory(CommandIdFactory())
        self._commands = {}
        with open(self._commands_json_path, 'r') as file:
            for i in json.load(file):
                self.add_command(self._command_factory.create(i['description'], i.get('keybinding'), i['exec'], i.get('meta')))



import os
import subprocess
from typing import Iterable, Optional, Dict, List

from libpalette.utils import remove_whitespace
from libpalette.configuration import Configuration
from libpalette.command import Command, CommandId, COMMAND_ID_SIZE

class Rofi:
    _configuration: Configuration

    def __init__(self, configuration: Configuration) -> None:
        self._configuration = configuration
    
    def show_palette(self, commands: Dict[CommandId, Command]) -> None:
        with open(self._configuration.rofi_script_input_path, 'wb') as file:
            file.write(self._get_rofi_input(commands))
        env_copy = os.environ.copy()
        env_copy['PALETTE_SOCKET'] = self._configuration.socket_path
        env_copy['PALETTE_ROFI_INPUT_PATH'] = self._configuration.rofi_script_input_path
        subprocess.Popen(['rofi', '-modi', f'palette:palette-rofi-script', '-show', 'palette'], env=env_copy)
    
    def _opposite_align(self, left: str, right: str, length: int) -> str:
        return left + ' ' * (length - len(right) - len(left)) + right
    
    def _format_keybinding(self, keybinding: str) -> str:
        if not self._configuration.use_modifier_icons:
            return f'[{keybinding}]'
        
        return ';'.join(
            ''.join(
                key.replace('alt', self._configuration.alt_modifier_icon)
                   .replace('ctrl', self._configuration.ctrl_modifier_icon)
                   .replace('super', self._configuration.super_modifier_icon)
                   .replace('shift', self._configuration.shift_modifier_icon)
                for key in chord.split('+')
            ) for chord in remove_whitespace(keybinding).split(';')
        )
    
    def _command_rofi_description(self, command: Command) -> str:
        return (
            command.description if command.keybinding is None
            else self._opposite_align(
                left = command.description,
                right = self._format_keybinding(command.keybinding),
                length = self._configuration.rofi_window_width
            ) if self._configuration.opposite_align_commands_and_shortcuts
            else f'{command.description} {self._format_keybinding(command.keybinding)}'
        )
    
    def _command_rofi_input(self, command: Command) -> str:
        metadata = f'\x1fmeta\x1f{command.metadata}' if command.metadata else ''
        return f'{self._command_rofi_description(command)}\0info\x1f{command.identifier}' + metadata

    def _properly_sorted(self, commands: Iterable[Command]) -> List[Command]:
        """Sorts the commands first by their last used time, and then by lexical order"""
        sorting_key = lambda command: (-command.last_used.toordinal(), command.description.casefold())
        return sorted(commands, key=sorting_key)

    def _get_rofi_input(self, commands: Dict[CommandId, Command]) -> bytes:
        return '\n'.join([
            "\0no-custom\x1ftrue",
            *map(self._command_rofi_input, self._properly_sorted(commands.values()))
        ]).encode()



import os
import socket
from typing import Callable

from libpalette.utils import try_remove_file
from libpalette.connection import Connection
from libpalette.connection_handler import ConnectionHandler

class UnixSocketServer:
    _socket: socket.socket
    _socket_path: str
    _connection_handler: ConnectionHandler

    def __init__(self, socket_path: str, connection_handler: ConnectionHandler) -> None:
        self._socket = socket.socket(socket.AF_UNIX)
        self._socket_path = socket_path
        self._connection_handler = connection_handler
        
    def __enter__(self) -> 'UnixSocketServer':
        try_remove_file(self._socket_path)
        self._socket.bind(self._socket_path)
        self._socket.listen()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        self._socket.close()
        try_remove_file(self._socket_path)

    def run(self) -> None:
        while True:
            with Connection(self._socket.accept()[0]) as connection:
                self._connection_handler.handle_connection(connection)



from typing import NamedTuple

class Configuration(NamedTuple):
    commands_path: str
    keybindings_backend: str = 'none'
    verbose: bool = False
    socket_path: str = '/tmp/palette_socket'
    rofi_script_input_path: str = '/tmp/palette-rofi-script_input'

    # TODO - more rofi customization
    opposite_align_commands_and_shortcuts: bool = True
    rofi_window_width: int  = 91

    # NOTE - modifier icons are not recommended as they are pretty ugly at the moment.
    use_modifier_icons:  bool = False
    ctrl_modifier_icon:  str  = '\ufb33'
    shift_modifier_icon: str  = '\ufb35'
    super_modifier_icon: str  = '\ufb32'
    alt_modifier_icon:   str  = '\ufb34'
